"use strict";

var _Object$defineProperty = require("@babel/runtime-corejs3/core-js-stable/object/define-property");

var _interopRequireDefault = require("@babel/runtime-corejs3/helpers/interopRequireDefault");

_Object$defineProperty(exports, "__esModule", {
  value: true
});

exports.decodeAddress = decodeAddress;
exports.decodeArgument = decodeArgument;
exports.decodeBool = decodeBool;
exports.decodeSignedInteger = decodeSignedInteger;
exports.decodeStaticBytes = decodeStaticBytes;
exports.decodeUnsignedInteger = decodeUnsignedInteger;
exports["default"] = decode;

var _startsWith = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/starts-with"));

var _padStart = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/pad-start"));

var _find = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/find"));

var _slice = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/slice"));

var _parseInt = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/number/parse-int"));

var _map = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/map"));

var _bignumber = _interopRequireDefault(require("bignumber.js"));

var _comprehension = _interopRequireDefault(require("comprehension"));

var _solidityUtils = require("./utils/solidity-utils");

var _numberRepresentations = require("./utils/number-representations");

var _converters = require("./utils/converters");

var _constants = require("../constants");

/** @param {string} value */
function checkValue(value) {
  if (typeof value !== 'string') throw new Error('decoding value is not a string');
  if (!/^[\da-fA-F]{64}$/.test(value)) throw new Error('decoding value is not a 32-byte in hex');
}
/**
 * @param {string} value
 * @return {boolean}
 */


function decodeBool(value) {
  if (value === (0, _comprehension["default"])(64, function () {
    return '0';
  }).join('')) return false;
  if (value === "".concat((0, _comprehension["default"])(63, function () {
    return '0';
  }).join(''), "1")) return true;
  throw new Error('unable to decode bool');
}
/**
 * @param {number} bitsCount
 * @param {string} value
 * @return {number|BigNumber}
 */


function decodeUnsignedInteger(bitsCount, value) {
  (0, _solidityUtils.checkIntegerSize)(bitsCount);
  checkValue(value);
  var result = new _bignumber["default"](value, 16);
  if (result.gte((0, _converters.toTwosPower)(bitsCount))) throw new Error("uint".concat(bitsCount, " overflow"));
  return bitsCount > 48 ? result : result.toNumber();
}
/**
 * @param {number} bitsCount
 * @param {string} value
 * @return {number|BigNumber}
 */


function decodeSignedInteger(bitsCount, value) {
  (0, _solidityUtils.checkIntegerSize)(bitsCount);
  checkValue(value);
  var twosComplementRepresentation = new _bignumber["default"](value, 16);

  if (twosComplementRepresentation.gte((0, _converters.toTwosPower)(bitsCount))) {
    throw new Error("int".concat(bitsCount, " overflow"));
  }

  var result = (0, _numberRepresentations.fromTwosComplementRepresentation)(twosComplementRepresentation, bitsCount);
  return bitsCount > 48 ? result : result.toNumber();
}
/**
 * @param {string} value
 * @return {string}
 */


function decodeAddress(value) {
  var _context, _context2, _context3;

  checkValue(value);
  if ((0, _startsWith["default"])(value).call(value, (0, _padStart["default"])(_context = '').call(_context, 24, '0')) && value[24] !== '0') return "0x".concat(value.substr(24));
  if (!(0, _startsWith["default"])(value).call(value, (0, _padStart["default"])(_context2 = '').call(_context2, 25, '0'))) throw new Error('first 100 bits are not zeros');

  var _13thByte = value.substr(24, 2);

  if (!/^0[01]$/.test(_13thByte)) throw new Error('13th byte is not in ["00", "01"]');
  var isContract = _13thByte === '01';
  var accountIndex = new _bignumber["default"](value.substr(26), 16);
  if (accountIndex.gte('2**32')) return "0x".concat((0, _padStart["default"])(_context3 = accountIndex.toString(16)).call(_context3, 40, '0'));
  return ['1', isContract ? _constants.PROTOCOL_OBJECT_TYPE_ID.CONTRACT : '2', accountIndex.toString(10)].join('.');
}
/**
 * @param {number} bytesCount
 * @param value
 * @return {Buffer}
 */


function decodeStaticBytes(bytesCount, value) {
  var _context4, _context5;

  (0, _solidityUtils.checkBytesCount)(bytesCount);
  checkValue(value);

  if ((0, _find["default"])(_context4 = (0, _slice["default"])(_context5 = value.split('')).call(_context5, bytesCount * 2)).call(_context4, function (hexChar) {
    return hexChar !== '0';
  })) {
    throw new Error("bytes".concat(bytesCount, " overflow"));
  }

  return Buffer.from(value.substr(0, bytesCount * 2), 'hex');
}
/**
 * @param {Array<string>} code
 * @param {number} from
 * @param {number} length
 * @param {import('../types/interfaces/Abi').SolType} type
 * @returns {{shift:number,res:Array<string>}}
 */


function decodeArray(code, from, length, type) {
  var shift = 0;
  var res = (0, _comprehension["default"])(length, function () {
    // eslint-disable-next-line no-shadow, no-use-before-define
    var _decodeArgument = decodeArgument(code, from + shift, type),
        elementShift = _decodeArgument.shift,
        res = _decodeArgument.res;

    shift += elementShift;
    return res;
  });
  return {
    shift: shift,
    res: res
  };
}
/**
 * @param {number} rawOffset
 * @returns {number}
 */


function getOffset(rawOffset) {
  var result = (0, _parseInt["default"])(rawOffset, 16);
  if (result % 32 !== 0) throw new Error('invalid offset');
  return result / 32;
}
/**
 * @param {Array<string>} code
 * @param {number} offset
 */


function decodeDynamicBytes(code, offset) {
  var length = (0, _parseInt["default"])(code[offset], 16);
  var hexCharsCount = length * 2;
  var fullCodes = (0, _comprehension["default"])(Math.floor(hexCharsCount / 64), function (i) {
    return code[offset + i + 1];
  });
  var postCode = hexCharsCount % 64 === 0 ? '' : code[offset + Math.ceil(hexCharsCount / 64)];

  if (hexCharsCount % 64 !== 0) {
    var _context6;

    if ((0, _find["default"])(_context6 = postCode.substr(hexCharsCount % 64).split('')).call(_context6, function (expectedZero) {
      return expectedZero !== '0';
    })) {
      throw new Error('non-zero bytes in code after full dynamic bytes');
    }

    postCode = postCode.substr(0, hexCharsCount % 64);
  }

  return Buffer.from(fullCodes.join('') + postCode, 'hex');
}
/**
 * @param {Array.<string>} code
 * @param {number} index
 * @param {SolType} type
 * @return {{ shift: number, res: * }}
 */


function decodeArgument(code, index, type) {
  if (type === 'bool') return {
    shift: 1,
    res: decodeBool(code[index])
  };
  if (type === 'address') return {
    shift: 1,
    res: decodeAddress(code[index])
  };
  if (type === 'bytes') return {
    shift: 1,
    res: decodeDynamicBytes(code, getOffset(code[index]))
  };
  if (type === 'string') return {
    shift: 1,
    res: decodeDynamicBytes(code, getOffset(code[index])).toString('utf8')
  };
  var uintMatch = type.match(/^uint(\d+)$/);

  if (uintMatch) {
    var bitsCount = (0, _parseInt["default"])(uintMatch[1], 10);
    return {
      shift: 1,
      res: decodeUnsignedInteger(bitsCount, code[index])
    };
  }

  var intMatch = type.match(/^int(\d+)$/);

  if (intMatch) {
    var _bitsCount = (0, _parseInt["default"])(intMatch[1], 10);

    return {
      shift: 1,
      res: decodeSignedInteger(_bitsCount, code[index])
    };
  }

  var staticBytesMatch = type.match(/^bytes(\d+)$/);

  if (staticBytesMatch) {
    var bytesCount = (0, _parseInt["default"])(staticBytesMatch[1], 10);
    return {
      shift: 1,
      res: decodeStaticBytes(bytesCount, code[index])
    };
  }

  var staticArrayMatch = type.match(/^(.+)\[(\d+)]$/);

  if (staticArrayMatch) {
    var typeOfMatch = staticArrayMatch[1];
    var length = (0, _parseInt["default"])(staticArrayMatch[2], 10);
    return decodeArray(code, index, length, typeOfMatch);
  }

  var dynamicArrayMatch = type.match(/^(.+)\[]$/);

  if (dynamicArrayMatch) {
    var _typeOfMatch = dynamicArrayMatch[1];
    var offset = getOffset(code[index]);

    var _length = (0, _parseInt["default"])(code[offset], 16);

    var _decodeArray = decodeArray(code, offset + 1, _length, _typeOfMatch),
        res = _decodeArray.res;

    return {
      shift: 1,
      res: res
    };
  }

  throw new Error("unknown type ".concat(type));
}
/**
 * @param {string} rawCode
 * @param {Array.<SolType>} types
 * @return {*|Array.<*>}
 */


function decode(rawCode, types) {
  if (rawCode.length % 64 !== 0) {
    if ((0, _slice["default"])(rawCode).call(rawCode, 0, 8) === '08c379a0') {
      var _context7;

      var errMessageLen = (0, _parseInt["default"])((0, _slice["default"])(rawCode).call(rawCode, 72, 136), 16);
      var errMessage = (0, _slice["default"])(_context7 = Buffer.from((0, _slice["default"])(rawCode).call(rawCode, 136), 'hex')).call(_context7, 0, errMessageLen).toString();
      throw new Error(errMessage);
    }

    throw new Error('length of code is not divisible by 32 bytes');
  }

  var code = (0, _comprehension["default"])({
    to: rawCode.length,
    step: 64
  }, function (i) {
    return rawCode.substr(i, 64);
  });
  var i = 0;
  /**
   * @param {SolType} type
   * @return {*}
   */

  var decodeType = function decodeType(type) {
    var _decodeArgument2 = decodeArgument(code, i, type),
        shift = _decodeArgument2.shift,
        res = _decodeArgument2.res;

    i += shift;
    return res;
  };

  var res = (0, _map["default"])(types).call(types, function (type) {
    return decodeType(type);
  });
  if (types.length === 1) return res[0];
  if (types.length === 0) return null;
  return res;
}